
#  Copyright 2020, 2023 Evandro Chagas Ribeiro da Rosa <evandro@quantuloop.com>
#  Copyright 2020, 2021 Rafael de Santiago <r.santiago@ufsc.br>
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from .base import set_quantum_execution_target, set_process_features
from .clib.kbw import (run_and_set_result, set_sim_mode_dense,
                       set_sim_mode_sparse, set_seed, set_dump_type)

__all__ = ['use_sparse', 'use_dense', 'set_seed', 'set_dump_type']


def use_sparse():
    """Set KBW Sparse as quantum execution target"""

    set_sim_mode_sparse()
    set_process_features(plugins=['pown'])
    set_quantum_execution_target(run_and_set_result)


def use_dense():
    """Set KBW Dense as quantum execution target"""

    set_sim_mode_dense()
    set_process_features(plugins=['pown'])
    set_quantum_execution_target(run_and_set_result)



from __future__ import annotations
#  Copyright 2020, 2023 Evandro Chagas Ribeiro da Rosa <evandro@quantuloop.com>
#  Copyright 2020, 2021 Rafael de Santiago <r.santiago@ufsc.br>
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ctypes import c_uint8, c_void_p, c_size_t, POINTER, c_int32
from os import environ
from os.path import dirname
from random import Random
import json
from .libket import JSON
from .wrapper import load_lib, from_u8_to_str, os_lib_name

DENSE = 0
SPARSE = 1

API_argtypes = {
    'kbw_run_and_set_result': ([c_void_p, c_int32], []),
    'kbw_run_serialized': ([POINTER(c_uint8), c_size_t, POINTER(c_uint8), c_size_t, c_int32, c_int32], [c_void_p]),   # pylint: disable=C0301
    'kbw_result_get': ([c_void_p], [POINTER(c_uint8), c_size_t]),
    'kbw_result_delete': ([c_void_p], []),
}


def kbw_path():
    """Get KBW path"""

    if "KBW_PATH" in environ:
        path = environ["KBW_PATH"]
    else:
        path = f'{dirname(__file__)}/libs/{os_lib_name("kbw")}'

    return path


API = load_lib('KBW', kbw_path(), API_argtypes, 'kbw_error_message')

SIM_MODE = None

RNG = Random()


def set_sim_mode_dense():
    """Set simulator mode to DENSE"""

    global SIM_MODE  # pylint: disable=W0603
    SIM_MODE = DENSE


def set_sim_mode_sparse():
    """Set simulator mode to SPARSE"""

    global SIM_MODE  # pylint: disable=W0603
    SIM_MODE = SPARSE


def set_seed(seed):
    """Initialize the simulator RNG"""

    global RNG  # pylint: disable=W0603
    RNG = Random(seed)


def set_dump_type(dump_type: str, shots: int | None = None):
    """Set the simulator dump type

    Args:
        dump_type: must be "vector", "probability", or "shots"
        shots: select the number of shots if ``dump_type`` is "shots"
    """

    if dump_type not in ["vector", "probability", "shots"]:
        raise ValueError('parameter "dump_type" must be "vector", "probability", or "shots"')

    environ['KBW_DUMP_TYPE'] = dump_type
    if shots is not None:
        environ['KBW_SHOTS'] = str(int(shots))


def run_and_set_result(process):
    """Execute quantum code from process"""

    environ['KBW_SEED'] = str(RNG.randint(0, (1 << 64) - 1))

    if SIM_MODE is None and 'KBW_MODE' in environ:
        sim_mode = environ['KBW_MODE'].upper()
        if sim_mode == 'DENSE':
            sim_mode = DENSE
        elif sim_mode == 'SPARSE':
            sim_mode = SPARSE
        else:
            raise RuntimeError(
                "undefined value for environment variable 'KBW_MODE', expecting 'DENSE' or 'SPARSE'")   # pylint: disable=C0301
    elif SIM_MODE is not None:
        sim_mode = SIM_MODE
    else:
        sim_mode = SPARSE

    API['kbw_run_and_set_result'](process, sim_mode)


def run_json(quantum_code, metrics, sim_mode: str = 'sparse'):
    """Call the simulator with a JSON quantum code"""

    quantum_code = json.dumps(quantum_code).encode()
    quantum_code_size = len(quantum_code)
    metrics = json.dumps(metrics).encode()
    metrics_size = len(metrics)

    sim_mode = {'SPARSE': SPARSE, 'DENSE': DENSE}[sim_mode.upper()]

    result = API['kbw_run_serialized'](
        (c_uint8 * quantum_code_size)(*quantum_code),
        quantum_code_size,
        (c_uint8 * metrics_size)(*metrics),
        metrics_size,
        JSON,
        sim_mode
    )

    data = json.loads(from_u8_to_str(*API['kbw_result_get'](result)))
    API['kbw_result_delete'](result)

    return data



from __future__ import annotations
#  Copyright 2020, 2023 Evandro Chagas Ribeiro da Rosa <evandro@quantuloop.com>
#  Copyright 2020, 2021 Rafael de Santiago <r.santiago@ufsc.br>
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
from typing import Any
from ket.base import (process_begin, process_end, process_top,
                      process_last, exec_quantum, set_process_features)
from ket.clib.libket import JSON
from ket.clib.wrapper import from_u8_to_str

__all__ = ['run', 'quantum_metrics', 'quantum_metrics_last',
           'quantum_code_last', 'quantum_code', 'quantum_exec_time',
           'quantum_exec_timeout', 'exec_quantum', 'set_process_features']


def quantum_metrics() -> dict[str, Any]:
    """Returns the metrics of the current process

    The return is a json serialization of a struct Metrics from Libket.
    See https://gitlab.com/quantum-ket/libket/-/blob/main/src/ir.rs for more details.
    """

    process_top().serialize_metrics(JSON)
    return json.loads(from_u8_to_str(*process_top().get_serialized_metrics()[:-1]))


def quantum_code() -> list[dict]:
    """Returns the quantum code of the current process

    The return is a json serialization of a struct CodeBlock list from Libket.
    See https://gitlab.com/quantum-ket/libket/-/blob/main/src/code_block.rs for more details.
    """

    process_top().serialize_quantum_code(JSON)
    return json.loads(from_u8_to_str(*process_top().get_serialized_quantum_code()[:-1]))


def quantum_metrics_last() -> dict[str, Any] | None:
    """Returns the metrics of the last executed process

    If no processes have been run, this function returns `None`.

    The return is a json serialization of a struct Metrics from Libket.
    See https://gitlab.com/quantum-ket/libket/-/blob/main/src/ir.rs for more details.
    """

    process_last().serialize_metrics(JSON)
    return json.loads(from_u8_to_str(*process_last().get_serialized_metrics()[:-1]))


def quantum_code_last() -> list[dict] | None:
    """Returns the quantum code of the last executed process

    The return is a json serialization of a struct CodeBlock list from Libket.
    See https://gitlab.com/quantum-ket/libket/-/blob/main/src/code_block.rs for more details.
    """

    process_last().serialize_quantum_code(JSON)
    return json.loads(from_u8_to_str(*process_last().get_serialized_quantum_code()[:-1]))


def quantum_exec_time() -> float | None:
    """Returns the quantum execution time in seconds of the last executed process

    If no processes have been run, this function returns `None`.
    """
    if process_last() is not None:
        return process_last().exec_time().value
    return None


def quantum_exec_timeout(timeout: int):
    """Set the quantum execution timeout

    Args:
        timeout: timeout in seconds.
    """

    process_top().set_timeout(timeout)


class run:  # pylint: disable=invalid-name
    """Execute in a new process

    Run the quantum operations in a new separated process.

    :Usage:

    .. code-block:: ket

        with run():
            ...

    """

    def __enter__(self):
        process_begin()

    def __exit__(self, type, value, tb):  # pylint: disable=redefined-builtin, invalid-name
        process_end()



#  Copyright 2020, 2023 Evandro Chagas Ribeiro da Rosa <evandro@quantuloop.com>
#  Copyright 2020, 2021 Rafael de Santiago <r.santiago@ufsc.br>
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import ast
import copy
# pylint: disable=missing-function-docstring, invalid-name


class ketpp (ast.NodeTransformer):
    """Ket interpreter preprocessor"""

    def __init__(self):
        self.id_count = 0
        self.while_begin = None
        self.while_end = None
        self.in_while = False

    def visit_If(self, node):

        if_id = self.id_count
        self.id_count += 1

        self.generic_visit(node)
        node_cp = copy.deepcopy(node)

        _test_name = '_ket_temp_if_test' + str(if_id)
        test_name = ast.Name(id=_test_name, ctx=ast.Load())

        end_name = '_ket_tmp_if_end' + str(if_id)
        else_name = None
        if node.orelse:
            else_name = '_ket_tmp_if_else' + str(if_id)
            if_begin = ast.Assign(
                targets=[
                    ast.Tuple(
                        elts=[
                            ast.Name(id=else_name, ctx=ast.Store()),
                            ast.Name(id=end_name, ctx=ast.Store())
                        ],
                        ctx=ast.Store()
                    )
                ],
                value=ast.Call(
                    func=ast.Name(id='_ket_if_else', ctx=ast.Load()),
                    args=[test_name],
                    keywords=[]
                )
            )
        else:
            if_begin = ast.Assign(
                targets=[
                    ast.Name(id=end_name, ctx=ast.Store()),
                ],
                value=ast.Call(
                    func=ast.Name(id='_ket_if', ctx=ast.Load()),
                    args=[test_name],
                    keywords=[]
                )
            )

        if_body = [if_begin, *node.body]

        if else_name is not None:
            else_call = ast.Expr(value=ast.Call(
                func=ast.Name(id='_ket_else', ctx=ast.Load()),
                args=[ast.Name(id=name, ctx=ast.Load())
                      for name in [else_name, end_name]],
                keywords=[]
            ))
            if_body.extend([else_call, *node.orelse])

        if_body.append(
            ast.Expr(value=ast.Call(
                func=ast.Name(id='_ket_next', ctx=ast.Load()),
                args=[ast.Name(id=end_name, ctx=ast.Load())],
                keywords=[]
            ))
        )

        test_assing = ast.Assign(
            targets=[ast.Name(id=_test_name, ctx=ast.Store())],
            value=node.test
        )

        type_check = ast.Call(
            func=ast.Name(id='_ket_is_future', ctx=ast.Load()),
            args=[test_name],
            keywords=[]
        )

        node_cp.test = test_name

        if_future = ast.If(
            test=type_check,
            body=if_body,
            orelse=[node_cp]
        )

        return [test_assing, if_future]

    def visit_While(self, node):  # pylint: disable=too-many-locals
        while_begin_save = self.while_begin
        while_end_save = self.while_end

        while_id = self.id_count
        self.id_count += 1

        begin_name = '_ket_tmp_while_begin' + str(while_id)
        loop_name = '_ket_tmp_while_loop' + str(while_id)
        end_name = '_ket_tmp_while_end' + str(while_id)
        else_name = None
        self.while_end = end_name
        if node.orelse:
            else_name = '_ket_tmp_while_else' + str(while_id)
            while_begin = ast.Assign(
                targets=[
                    ast.Tuple(
                        elts=[ast.Name(id=name, ctx=ast.Store()) for name in [
                            begin_name, loop_name, else_name, end_name]],
                        ctx=ast.Store()
                    )
                ],
                value=ast.Call(
                    func=ast.Name(id='_ket_while_else', ctx=ast.Load()),
                    args=[],
                    keywords=[]
                )
            )
        else:
            while_begin = ast.Assign(
                targets=[
                    ast.Tuple(
                        elts=[ast.Name(id=name, ctx=ast.Store())
                              for name in [begin_name, loop_name, end_name]],
                        ctx=ast.Store()
                    )
                ],
                value=ast.Call(
                    func=ast.Name(id='_ket_while', ctx=ast.Load()),
                    args=[],
                    keywords=[]
                )
            )

        test_name = '_ket_while_test' + str(while_id)
        test_assing = ast.Assign(
            targets=[ast.Name(id=test_name, ctx=ast.Store())],
            value=node.test
        )

        type_check = ast.Call(
            func=ast.Name(id='_ket_is_future', ctx=ast.Load()),
            args=[ast.Name(id=test_name, ctx=ast.Load())],
            keywords=[]
        )

        body_call = ast.Expr(value=ast.Call(
            func=ast.Name(id='_ket_while_body', ctx=ast.Load()),
            args=[
                ast.Name(id=test_name, ctx=ast.Load()),
                ast.Name(id=loop_name, ctx=ast.Load()),
                ast.Name(
                    id=else_name if else_name is not None else end_name, ctx=ast.Load())
            ],
            keywords=[]
        ))

        loop_call = ast.Expr(value=ast.Call(
            func=ast.Name(id='_ket_loop', ctx=ast.Load()),
            args=[
                ast.Name(id=begin_name, ctx=ast.Load()),
                ast.Name(
                    id=else_name if else_name is not None else end_name, ctx=ast.Load())
            ],
            keywords=[]
        ))

        else_call = [ast.Expr(value=ast.Call(
            func=ast.Name(id='_ket_next', ctx=ast.Load()),
            args=[ast.Name(id=end_name, ctx=ast.Load())],
            keywords=[]
        ))] if else_name is not None else []

        node_cp = copy.deepcopy(node)
        in_while_save = self.in_while
        self.in_while = True
        self.generic_visit(node)
        self.in_while = in_while_save

        self.generic_visit(node_cp)

        ndone_name = '_ket_tmp_while_ndone' + str(while_id)

        if_not_future = ast.If(
            test=ast.Name(id=test_name, ctx=ast.Load()),
            body=[
                ast.Assign(
                    targets=[ast.Name(id=ndone_name, ctx=ast.Store())],
                    value=ast.Constant(value=True)
                ),
                ast.While(
                    test=ast.Name(id=ndone_name, ctx=ast.Load()),
                    body=[*node_cp.body,
                          ast.If(
                              test=ast.UnaryOp(
                                  op=ast.Not(), operand=node_cp.test),
                              body=[
                                  ast.Assign(
                                      targets=[
                                          ast.Name(id=ndone_name, ctx=ast.Store())],
                                      value=ast.Constant(value=False)
                                  )
                              ],
                              orelse=[]
                          )
                          ],
                    orelse=[]
                ),
                ast.If(
                    test=ast.UnaryOp(
                        op=ast.Not(),
                        operand=ast.Name(id=ndone_name, ctx=ast.Load())
                    ),
                    body=node_cp.orelse if node_cp.orelse else [ast.Pass()],
                    orelse=[]
                )
            ],
            orelse=node_cp.orelse
        )

        if_future = ast.If(
            test=type_check,
            body=[body_call, *node.body, loop_call, *else_call],
            orelse=[if_not_future]
        )

        self.while_begin = while_begin_save
        self.while_end = while_end_save

        return [while_begin, test_assing, if_future]

    def visit_Break(self, node):
        return ast.Expr(value=ast.Call(
            func=ast.Name(id='_ket_goto', ctx=ast.Load()),
            args=[ast.Name(id=self.while_end, ctx=ast.Load())],
            keywords=[]
        )) if self.in_while else node

    def visit_Continue(self, node):
        return ast.Expr(value=ast.Call(
            func=ast.Name(id='_ket_goto', ctx=ast.Load()),
            args=[ast.Name(id=self.while_begin, ctx=ast.Load())],
            keywords=[]
        )) if self.in_while else node



from ket.gates import *
from ket.standard import ctrl, adj
from ket.lib import dump_matrix
from math import sqrt, pi, cos, sin
from cmath import exp
from ket import kbw

kbw.use_sparse()

gate_matrix = {
    I: [[1, 0],
        [0, 1]],

    X: [[0, 1],
        [1, 0]],

    Y: [[0, -1j],
        [1j,  0]],

    Z: [[1,   0],
        [0, -1]],

    H: [[1/sqrt(2),  1/sqrt(2)],
        [1/sqrt(2), -1/sqrt(2)]],

    S: [[1,   0],
        [0,  1j]],

    SD: [[1,    0],
         [0,  -1j]],

    T: [[1,             0],
        [0,  exp(1j*pi/4)]],

    TD: [[1,              0],
         [0,  exp(-1j*pi/4)]]
}

param_gate_matrix = {
    phase: lambda theta: [[1,               0],
                          [0, exp(1j*theta)]],

    RX: lambda theta: [[cos(theta/2), -1j*sin(theta/2)],
                       [-1j*sin(theta/2), cos(theta/2)]],

    RY: lambda theta: [[cos(theta/2), -sin(theta/2)],
                       [sin(theta/2),  cos(theta/2)]],

    RZ: lambda theta: [[exp(-1j*theta/2), 0],
                       [0,  exp(1j*theta/2)]]
}


def make_ctrl_gate(mat):
    return [[1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0]+mat[0],
            [0, 0]+mat[1]]


def make_adj_gate(mat):
    return [[complex(mat[j][i]).conjugate() for j in range(len(mat))] for i in range(len(mat[0]))]


def tau_range(parts=100):
    part = 2*pi/parts
    step = 0
    while step < 2*pi:
        yield step
        step += part
    return 2*pi


def is_equal(a, b, epsilon=1e-10):
    return all(all(abs(aj-bj) < epsilon for aj, bj in zip(ai, bi)) for ai, bi in zip(a, b))


def test_gates():
    for gate in [I, X, Y, Z, H, S, SD, T, TD]:
        for sim in [kbw.use_sparse, kbw.use_dense]:
            sim()
            print(gate, sim)
            assert is_equal(dump_matrix(gate), gate_matrix[gate])
            assert is_equal(dump_matrix(ctrl(0, gate, 1), 2),
                            make_ctrl_gate(gate_matrix[gate]))
            assert is_equal(dump_matrix(adj(gate)),
                            make_adj_gate(gate_matrix[gate]))
            assert is_equal(dump_matrix(ctrl(0, adj(gate), 1), 2),
                            make_ctrl_gate(make_adj_gate(gate_matrix[gate])))


def test_param_gates():
    for gate in [phase, RX, RY, RZ]:
        for sim in [kbw.use_sparse, kbw.use_dense]:
            sim()
            print(gate, sim)

            assert all(is_equal(dump_matrix(gate(theta)),
                                param_gate_matrix[gate](theta)) for theta in tau_range())
            assert all(is_equal(dump_matrix(ctrl(0, gate(theta), 1), 2),
                                make_ctrl_gate(param_gate_matrix[gate](theta))) for theta in tau_range())
            assert all(is_equal(dump_matrix(adj(gate(theta))),
                                make_adj_gate(param_gate_matrix[gate](theta))) for theta in tau_range())
            assert all(is_equal(dump_matrix(ctrl(0, adj(gate(theta)), 1), 2),
                                make_ctrl_gate(make_adj_gate(param_gate_matrix[gate](theta)))) for theta in tau_range())



from __future__ import annotations
#  Copyright 2020, 2023 Evandro Chagas Ribeiro da Rosa <evandro@quantuloop.com>
#  Copyright 2020, 2021 Rafael de Santiago <r.santiago@ufsc.br>
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from inspect import getsource
from ast import fix_missing_locations, parse
from os import path, PathLike
from types import ModuleType
from typing import Callable
from .preprocessor.transformer import ketpp

__all__ = ['import_ket', 'from_import_ket']


def _import_ket(source: PathLike):
    with open(source, 'r', encoding="utf-8") as file:
        tree = parse(file.read(), source)

    preprocessor = ketpp()

    preprocessor.visit(tree)
    fix_missing_locations(tree)

    return compile(tree, source, 'exec', optimize=2)

# pylint: disable=exec-used


def _import_globals_ket(source: PathLike, global_vars):
    exec(_import_ket(source), global_vars)


BUILDINS = \
    """
from ket import *
from ket.base import label, branch, jump
from ket.preprocessor.statements import *
"""


def import_ket(source: PathLike) -> ModuleType:
    """Import Ket file

    Import a Ket source file as a Python module.

    Args:
        source: Ket source file.
    """

    _, file_name = path.split(source)
    module = ModuleType(file_name[:-4])

    obj = compile(BUILDINS, '<ket build-in functions>', 'exec', optimize=2)
    exec(obj, module.__dict__)

    _import_globals_ket(source, module.__dict__)
    return module


def from_import_ket(source: PathLike, *names: list[str]) -> tuple:
    """Import names from Ket file

    Args:
        source: Ket source file.
        names: Names to import from ``source``.
    """

    module = import_ket(source)
    return tuple(module.__dict__[name] for name in names)


def code_ket(func: Callable) -> Callable:
    """Parse as Ket function

    Decorator to parse a Python function as a Ket Function.

    Warning:
        * Do not use this decorator in a .ket file.
        * This decorator do not work with function defined inside
          a Jupyter Notebook or the Python interpreter.
        * This decorator is not part of Ket's preamble.

    :Usage:

    .. code-block:: python

        from ket import code_ket

        @code_ket
        def func(a : quant, b : quant, c : quant):
            m0 = measure(a)
            m1 = measure(b)
            if m1 == 1: # need @code_ket to properly execute
                x(c)
            if m0 == 1: # need @code_ket to properly execute
                z(c)
    """

    if '__in_ket__' in globals() and globals()['__in_ket__']:
        return func

    in_ket = f"__in_ket__{func.__name__}__"
    if in_ket in func.__globals__ and func.__globals__[in_ket]:
        return func

    doc = func.__doc__
    annotations = func.__annotations__

    buildins_more = BUILDINS + '\n' + in_ket + ' = True\n'
    buildins_obj = compile(
        buildins_more, '<ket build-in functions>', 'exec', optimize=2)
    exec(buildins_obj, func.__globals__)

    tree = parse(getsource(func), '<function ' + func.__name__ + '>')
    preprocessor = ketpp()

    preprocessor.visit(tree)
    fix_missing_locations(tree)

    obj = compile(tree, '<code_ket function ' + func.__name__ + '>', 'exec', optimize=2)
    exec(obj, func.__globals__)
    func.__globals__[func.__name__].__doc__ = doc
    func.__globals__[func.__name__].__annotations__ = annotations

    return func.__globals__[func.__name__]



from __future__ import annotations
#  Copyright 2020, 2023 Evandro Chagas Ribeiro da Rosa <evandro@quantuloop.com>
#  Copyright 2020, 2021 Rafael de Santiago <r.santiago@ufsc.br>
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from .base import plugin, quant, base_X

__all__ = ['plugin', 'pown']


def pown(a: int, x: quant, N: int) -> quant:  # pylint: disable=invalid-name
    r"""Apply a modular exponentiation in a superposition

    .. math::

        \left| x \right> \left| 1 \right> \rightarrow  \left| x \right> \left| a^x\; \text{mod} \, N \right>

    .. note::

        Plugin availability depends on the quantum execution target.

    Args:
        a: :math:`a`.
        x: :math:`x`.
        N: :math:`N`.

    :return: :class:`~ket.base.quant` with the operation result.
    """  # pylint: disable=line-too-long

    ret = quant(N.bit_length())
    base_X(ret[-1])

    plugin('pown', f'{a} {N}', x + ret)

    return ret
